#  musicview, (re)discover your music library.
#  Copyright (C) 2018 Peijun Ma
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import signal
from sqlite3 import connect
from subprocess import DEVNULL, Popen
from threading import Condition, RLock, Thread
from time import sleep

from .db import next_song
from .misc import format_time


class Player:
    """
    Music player class
    """

    def __init__(self, data, name, ffplay, conn, controls, stdscr):
        """
        Args:
            data: path to music library
            name: name of the music library
            ffplay: ffplay binary
            conn: database connection
            controls: curses controls keymap
            stdscr: curses screen
        """
        self.ffplay = ffplay
        self.controls = controls
        self.data = data
        self.name = name
        self.conn = conn
        self.stdscr = stdscr
        self.height, self.width = stdscr.getmaxyx()

        self.cur_song = None
        self.cur_fav = False
        self.cur_count = 0
        self.time_elapsed = 0

        self.stopped = False
        self.paused = True

        self.playing_proc = None
        self.ui_thread = None

        self.db_lock = RLock()
        self.cv = Condition()

    def _play(self) -> Popen:
        """
        Return a subprocess to play the current song
        Returns:
            a subprocess to play the current song
        """
        return Popen(
            [self.ffplay, '-nodisp', '-autoexit', self.cur_song.path],
            stderr=DEVNULL
        )

    def ui(self):
        """
        UI control, meant to be ran in another thread
        """
        conn = connect(str(self.data / self.name))
        while True:
            cmd = self.controls.get(self.stdscr.getkey())
            if cmd == 'play/pause':
                self.toggle()
            elif cmd == 'skip':
                self.play_next()
            elif cmd == 'toggle favourite':
                self.favourite(conn)
            elif cmd == 'quit':
                self.stopped = True
                self.playing_proc.kill()
                self.stdscr.clear()
                return

    def progress(self):
        """
        Progress bar control, meant to be ran in another thread
        """
        while True:
            if self.stopped:
                return
            with self.cv:
                while self.paused:
                    self.cv.wait()
                self.display()
                sleep(1)
                self.time_elapsed += 1
                if self.time_elapsed > self.cur_song.length:
                    self.play_next()

    def play_next(self):
        """
        Play the next song
        """
        self.playing_proc.kill()

    def toggle(self):
        """
        Toggle play/pause status
        """
        if self.paused:
            self.playing_proc.send_signal(signal.SIGCONT)
            with self.cv:
                self.cv.notify()
        else:
            self.playing_proc.send_signal(signal.SIGSTOP)
        self.paused = not self.paused
        self.display()

    def start(self):
        """
        Start the music player
        """
        ui = Thread(target=self.ui, name='ui', daemon=True)
        self.ui_thread = ui
        ui.start()
        progress = Thread(target=self.progress, name='progress', daemon=True)
        progress.start()
        while True:
            self.paused = True
            if self.stopped:
                return
            with self.cv:
                self.time_elapsed = 0
            with self.db_lock:
                self.cur_song, (self.cur_fav, self.cur_count) = next_song(self.conn)
                self.cur_count += 1
            self.display()
            with self._play() as proc:
                self.playing_proc = proc
                self.paused = False
                with self.cv:
                    self.cv.notify()

    def favourite(self, conn):
        """
        Toggle the favourite status of the current song
        Args:
            conn: connection to the sqlite database
        """
        with self.db_lock:
            conn.execute(
                'UPDATE library SET favourite=? WHERE path=?',
                (not self.cur_fav, self.cur_song.path)
            )
            conn.commit()
        self.cur_fav = not self.cur_fav
        self.display()

    def display(self):
        """
        Display text on screen
        """
        self.stdscr.clear()
        length = self.cur_song.length
        total_time = self.cur_song.format_time()
        cur_time = format_time(self.time_elapsed)
        self.stdscr.addstr(0, 1, '[paused]' if self.paused else '[playing]')
        self.stdscr.addstr(
            1, 1, '{}/{}'.format(cur_time, total_time)
        )
        bar_len = self.width - 4
        prog = int(bar_len * self.time_elapsed // length) - 1
        if prog < 0:
            prog = 0
        empty = bar_len - prog - 1
        self.stdscr.addstr(2, 1, '|{}{}{}|'.format(prog * '=', '>', empty * ' '))
        self.stdscr.addstr(
            3, 1,
            '\n '.join(
                self.cur_song.format() +
                ['Favourite: {}'.format(bool(self.cur_fav)),
                 'Play count: {}\n'.format(self.cur_count)]
            )
        )
        self.stdscr.refresh()



#  musicview, (re)discover your music library.
#  Copyright (C) 2018 Peijun Ma
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import Iterable, namedtuple
from subprocess import DEVNULL, PIPE, run

from mutagen import File, MutagenError

from .misc import format_time


class MetaData(namedtuple('MetaData', ['path', 'title', 'genre', 'artist', 'album', 'length'])):
    @classmethod
    def empty(cls, path):
        """
        Return an empty MetaData entry
        Args:
            path: path to the song

        Returns:
            empty MetaData entry
        """
        return cls(path, None, None, None, None, None)

    def format_time(self):
        """
        Format self.length into human readable string
        Returns:
            The time formatted
        """
        if self.length:
            return format_time(self.length)
        else:
            return None

    def format(self) -> list:
        """
        Format self into a pretty printable target,
        with one line per list
        Returns:
            the formatted text
        """
        title = self.title or self.path
        length = self.format_time()
        lst = list(self[1:])
        lst[0] = title
        lst[-1] = length
        return ['{}: {}'.format(s, v) for s, v in
                zip(['Title', 'Genre', 'Artist', 'Album', 'Length'], lst)
                if v]


def tag_to_str(tag):
    """
    Function to turn a tag into a string.
    Args:
        tag: the tag value
    Returns: the tag as a string
    """
    if isinstance(tag, str):
        res = tag.strip()
    elif isinstance(tag, Iterable):
        res = ','.join(tag).strip()
    else:
        res = None
    if isinstance(res, str):
        return res or None
    return None


def song_metadata(ffmpeg, path: str):
    """
    Get a song's metadata
    Args:
        ffmpeg: ffmpeg binary
        path: path to the song
    Returns:
        The song's metadata
    """

    try:
        tags = File(path, easy=True)
    except MutagenError:
        res = MetaData.empty(path)
    else:
        if not tags:
            res = MetaData.empty(path)
        else:
            get = lambda t, s: t.get(s, t.get(s.upper()))
            title = tag_to_str(get(tags, 'title'))
            genre = tag_to_str(get(tags, 'genre'))
            artist = tag_to_str(get(tags, 'artist'))
            album = tag_to_str(get(tags, 'album'))
            length = tags.info.length
            res = MetaData(path, title, genre, artist, album, length)

    if res.length is None:  # Couldn't read length from mutagen, fallback to ffmpeg
        proc = run([ffmpeg, '-i', path], stdout=DEVNULL, stderr=PIPE)
        err = proc.stderr.decode().splitlines()
        for line in err:
            line = line.strip().lower()
            if line.startswith('duration'):
                try:
                    time = line.split()[1].rstrip(',')
                    hour, minutes, seconds = time.split(':')
                    total = 60 * 60 * int(hour) + 60 * int(minutes) + float(seconds)
                except Exception:
                    return None
                else:
                    return MetaData(*res[:-1], total)
        return None
    else:
        return res



#  musicview, (re)discover your music library.
#  Copyright (C) 2018 Peijun Ma
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

__title__ = 'musicview'
__description__ = '(re)discover your music library.'
__url__ = 'https://github.com/MaT1g3R/musicview'
__version__ = '0.2.0'
__author__ = 'Peijun Ma'
__author_email__ = 'peijun.ma@protonmail.com'
__license__ = 'GPLv3.0+'
__copyright__ = 'Copyright (C) 2018 Peijun Ma'



"""Python module which parses and emits TOML.

Released under the MIT license.
"""

from toml import encoder
from toml import decoder

__version__ = "0.9.3"
__spec__ = "0.4.0"

load = decoder.load
loads = decoder.loads
TomlDecoder = decoder.TomlDecoder

dump = encoder.dump
dumps = encoder.dumps

TomlDecodeError = decoder.TomlDecodeError



#  musicview, (re)discover your music library.
#  Copyright (C) 2018 Peijun Ma
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import sys
import urllib.error
import urllib.request
from curses import wrapper
from functools import partial
from os import getenv
from pathlib import Path
from shutil import which
from sqlite3 import connect

import click

from .__version__ import __title__, __version__
from .db import init_db, update_db
from .player import Player

sys.path.append(str(Path(__file__).parent.parent / 'toml'))
import toml

DEFAULT_CONFIG_HOME = Path.home() / '.musicview'
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
CONFIG_ENVAR = 'MUSICVIEW_CONFIG_HOME'
CONF_FILE = 'musicview.toml'

print = click.echo
fprint = partial(click.echo, file=sys.stderr)


class Ctx:
    """
    Context for the cli
    """

    def __init__(self):
        ffplay = which('ffplay')
        ffmpeg = which('ffmpeg')
        if not ffplay or not ffmpeg:
            print('ffmpeg/ffplay not found!', file=sys.stderr)
            exit(1)
        self.ffplay = ffplay
        self.ffmpeg = ffmpeg
        self.config_home = Path(getenv(CONFIG_ENVAR, DEFAULT_CONFIG_HOME))
        if not (self.config_home / CONF_FILE).is_file():
            self.setup()
        self.config = toml.loads((self.config_home / CONF_FILE).read_text())
        if self.config['general']['check for updates']:
            check_update()

    def setup(self):
        """Prompt the user to do some initial setup"""
        click.confirm(
            f"It seems like it's your first time using {__title__}!\n"
            f'Would you like to store configuration and metadata under'
            f' {self.config_home}?\n'
            f'If not, you can set the {CONFIG_ENVAR} environment variable'
            f' to change the location.',
            abort=True
        )
        try:
            self.config_home.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            exit(e)

        check_update = click.confirm(
            'Would you like to automatically check for updates?'
        )
        default_config = {
            'general': {
                'check for updates': check_update,
            },
            'player control': {
                'play/pause': 'p',
                'skip': '>',
                'toggle favourite': 'f',
                'quit': 'q',
            },
            'library paths': {}
        }
        self.dump_config(default_config)
        print(f'Configuration file has been generated at '
              f'{self.config_home / CONF_FILE}')

    def dump_config(self, cfg):
        """
        Dump a config dict
        Args:
            cfg: The config to dump
        """
        with (self.config_home / CONF_FILE).open('w+') as f:
            try:
                toml.dump(cfg, f)
            except OSError as e:
                exit(e)

    def get_libpath(self, name):
        """
        Get library path by name
        Args:
            name: name of the library

        Returns:
            path to the library
        """
        return Path(self.config['library paths'][name])

    def library_exists(self, name):
        """
        Check if a library exists
        Args:
            name: The name for the library

        Returns:
            if the library exists
        """
        return name in map(lambda p: p.name.rstrip('.db'), self.config_home.iterdir()) \
               and name in self.config['library paths']

    def delete_lib(self, name):
        """
        Delete a music library
        Args:
            name: Name of the music library
        """
        del self.config['library paths'][name]
        self.dump_config(self.config)

    def new_lib(self, name, path):
        """
        Add a new music library
        Args:
            name: name of the library
            path: path of the library
        """
        assert name not in self.config['library paths']
        self.config['library paths'][name] = path
        self.dump_config(self.config)

    def get_conn(self, name):
        """
        Get a sqlite3 connection
        Args:
            name: name of the library

        Returns:
            the connection
        """
        return connect(str(self.config_home / f'{name}.db'))


pass_context = click.make_pass_decorator(Ctx, ensure=True)


def check_update():
    """Check for updates"""
    url = ('https://raw.githubusercontent.com'
           '/MaT1g3R/musicview/master/musicview/__version__.py')
    try:
        with urllib.request.urlopen(url) as resp:
            text = resp.read().decode()
    except urllib.error.URLError:
        print('Could not check for updates, '
              'are you connected to the internet?')
    else:
        regex = re.compile(r"__version__[\s]*=[\s]*\'(.*)\'")
        head_version = regex.findall(text)[0]
        t = lambda s: tuple(map(int, s.split('.')))
        if t(head_version) > t(__version__):
            print(f'New version ({head_version}) available!')


def play_music(data, name, conn, ffplay, controls, stdscr):
    """
    Play some music!
    Args:
        data: path to the dirctory containg database files
        name: name of the music library
        conn: database connection
        ffplay: ffplay binary
        controls: curses controls keymap
        stdscr: curses screen
    """
    stdscr.clear()
    player = Player(data, name, ffplay, conn, controls, stdscr)
    player.start()


@click.group(context_settings=CONTEXT_SETTINGS)
@pass_context
def cli(ctx):
    """musicview, (re)discover your music library"""
    pass


@click.command(name='list')
@pass_context
def list_(ctx):
    """List existing music libraries"""
    names = ctx.config['library paths']
    if names:
        for name, path in names.items():
            print(f'{name} (at {path})')
    else:
        print('There are currently no music libraries!')


@click.command()
@click.argument('name')
@pass_context
def play(ctx, name):
    """Start playing music"""
    controls = {val: key for key, val in ctx.config['player control'].items()}
    with ctx.get_conn(name) as conn:
        wrapper(partial(play_music, ctx.config_home, name, conn, ctx.ffplay, controls))


@click.command()
@click.argument('name')
@pass_context
def delete(ctx, name: str):
    """Delete a music library"""
    to_del = ctx.config_home / f'{name}.db'
    click.confirm(f'Delete {name}?', abort=True)
    ctx.delete_lib(name)
    print(f'{name} deleted!')
    try:
        to_del.unlink()
    except OSError as e:
        fprint(e)


@click.command()
@click.argument('name')
@pass_context
def update(ctx, name: str):
    """Update an existing music library"""
    if not ctx.library_exists(name):
        exit(f'Library "{name}" does not exist!')
    with ctx.get_conn(name) as conn:
        update_db(ctx.get_libpath(name), conn, ctx.ffmpeg, ctx.ffplay)


@click.command(short_help='Create a new music library')
@click.argument('name')
@click.argument('path', type=click.Path(exists=True, file_okay=False))
@pass_context
def new(ctx, name, path):
    """
    Create a new music library

    Arguments:

        NAME: Name of the new music library

        PATH: Path of the music library
    """
    if ctx.library_exists(name):
        exit(f'Library with name {name} already exists!')
    path = Path(path).resolve()
    init_db(
        path, ctx.config_home / f'{name}.db', ctx.ffmpeg, ctx.ffplay
    )
    ctx.new_lib(name, str(path))


cli.add_command(list_)
cli.add_command(new)
cli.add_command(update)
cli.add_command(delete)
cli.add_command(play)



#!/usr/bin/env python3
#  musicview, (re)discover your music library.
#  Copyright (C) 2018 Peijun Ma
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import sys
from pathlib import Path
from shlex import split
from subprocess import run

from setuptools import setup

HERE = Path(__file__).parent
sys.path.append(str(HERE / 'toml'))
import toml

if sys.argv[-1] == 'publish':
    run(split('python setup.py sdist bdist_wheel'))
    run(split('twine upload dist/*'))
    exit()

pipfile = toml.loads((HERE / 'Pipfile').read_text())
lock = json.loads((HERE / 'Pipfile.lock').read_text())
readme = (HERE / 'README.rst').read_text()

about = {}
exec((HERE / 'musicview' / '__version__.py').read_text(), about)

reqs = [
    key + val['version'].replace('==', '>=')
    for key, val in lock['default'].items()
    if key in pipfile['packages']
]

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=readme,
    author=about['__author__'],
    author_email=about['__author_email__'],
    license=about['__license__'],
    url=about['__url__'],
    py_modules=[
        'main',
    ],
    packages=[
        'musicview',
        'toml/toml',
    ],
    install_requires=reqs,
    package_data={'': ['README.rst', 'LICENSE', 'Pipfile', 'Pipfile.lock']},
    include_package_data=True,
    python_requires=">=3.6",
    entry_points='''
        [console_scripts]
        musicview=main:main    
    ''',
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Console :: Curses',
        'Intended Audience :: End Users/Desktop',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Multimedia :: Sound/Audio :: Players',
        'Topic :: Multimedia :: Sound/Audio :: Players :: MP3'
    ),
)



import datetime
import re
import sys

from toml.decoder import InlineTableDict

if sys.version_info >= (3,):
    unicode = str


def dump(o, f):
    """Writes out dict as toml to a file

    Args:
        o: Object to dump into toml
        f: File descriptor where the toml should be stored

    Returns:
        String containing the toml corresponding to dictionary

    Raises:
        TypeError: When anything other than file descriptor is passed
    """

    if not f.write:
        raise TypeError("You can only dump an object to a file descriptor")
    d = dumps(o)
    f.write(d)
    return d


def dumps(o, encoder=None):
    """Stringifies input dict as toml

    Args:
        o: Object to dump into toml

        preserve: Boolean parameter. If true, preserve inline tables.

    Returns:
        String containing the toml corresponding to dict
    """

    retval = ""
    if encoder is None:
        encoder = TomlEncoder()
    addtoretval, sections = encoder._dump_sections(o, "")
    retval += addtoretval
    while sections != {}:
        newsections = {}
        for section in sections:
            addtoretval, addtosections = encoder._dump_sections(
                sections[section], section)

            if addtoretval or (not addtoretval and not addtosections):
                if retval and retval[-2:] != "\n\n":
                    retval += "\n"
                retval += "[" + section + "]\n"
                if addtoretval:
                    retval += addtoretval
            for s in addtosections:
                newsections[section + "." + s] = addtosections[s]
        sections = newsections
    return retval


class TomlEncoder(object):

    def __init__(self, preserve=False):
        self.preserve = preserve

    def _dump_sections(self, o, sup):
        retstr = ""
        if sup != "" and sup[-1] != ".":
            sup += '.'
        retdict = o.__class__()
        arraystr = ""
        for section in o:
            section = unicode(section)
            qsection = section
            if not re.match(r'^[A-Za-z0-9_-]+$', section):
                if '"' in section:
                    qsection = "'" + section + "'"
                else:
                    qsection = '"' + section + '"'
            if not isinstance(o[section], dict):
                arrayoftables = False
                if isinstance(o[section], list):
                    for a in o[section]:
                        if isinstance(a, dict):
                            arrayoftables = True
                if arrayoftables:
                    for a in o[section]:
                        arraytabstr = "\n"
                        arraystr += "[[" + sup + qsection + "]]\n"
                        s, d = self._dump_sections(a, sup + qsection)
                        if s:
                            if s[0] == "[":
                                arraytabstr += s
                            else:
                                arraystr += s
                        while d != {}:
                            newd = {}
                            for dsec in d:
                                s1, d1 = self._dump_sections(d[dsec], sup +
                                                             qsection + "." +
                                                             dsec)
                                if s1:
                                    arraytabstr += ("[" + sup + qsection +
                                                    "." + dsec + "]\n")
                                    arraytabstr += s1
                                for s1 in d1:
                                    newd[dsec + "." + s1] = d1[s1]
                            d = newd
                        arraystr += arraytabstr
                else:
                    if o[section] is not None:
                        retstr += (qsection + " = " +
                                   unicode(_dump_value(o[section])) + '\n')
            elif self.preserve and isinstance(o[section], InlineTableDict):
                retstr += (qsection + " = " + _dump_inline_table(o[section]))
            else:
                retdict[qsection] = o[section]
        retstr += arraystr
        return (retstr, retdict)


def _dump_inline_table(section):
    """Preserve inline table in its compact syntax instead of expanding
    into subsection.

    https://github.com/toml-lang/toml#user-content-inline-table
    """
    retval = ""
    if isinstance(section, dict):
        val_list = []
        for k, v in section.items():
            val = _dump_inline_table(v)
            val_list.append(k + " = " + val)
        retval += "{ " + ", ".join(val_list) + " }\n"
        return retval
    else:
        return unicode(_dump_value(section))


def _dump_value(v):
    dump_funcs = {
        str: lambda: _dump_str(v),
        unicode: lambda: _dump_str(v),
        list: lambda: _dump_list(v),
        bool: lambda: unicode(v).lower(),
        float: lambda: _dump_float(v),
        datetime.datetime: lambda: v.isoformat().replace('+00:00', 'Z'),
    }
    # Lookup function corresponding to v's type
    dump_fn = dump_funcs.get(type(v))
    # Evaluate function (if it exists) else return v
    return dump_fn() if dump_fn is not None else v


def _dump_str(v):
    v = "%r" % v
    if v[0] == 'u':
        v = v[1:]
    singlequote = v.startswith("'")
    v = v[1:-1]
    if singlequote:
        v = v.replace("\\'", "'")
        v = v.replace('"', '\\"')
    v = v.split("\\x")
    while len(v) > 1:
        i = -1
        if not v[0]:
            v = v[1:]
        v[0] = v[0].replace("\\\\", "\\")
        # No, I don't know why != works and == breaks
        joinx = v[0][i] != "\\"
        while v[0][:i] and v[0][i] == "\\":
            joinx = not joinx
            i -= 1
        if joinx:
            joiner = "x"
        else:
            joiner = "u00"
        v = [v[0] + joiner + v[1]] + v[2:]
    return unicode('"' + v[0] + '"')


def _dump_list(v):
    retval = "["
    for u in v:
        retval += " " + unicode(_dump_value(u)) + ","
    retval += "]"
    return retval


def _dump_float(v):
    return "{0:.16}".format(v).replace("e+0", "e+").replace("e-0", "e-")



#  musicview, (re)discover your music library.
#  Copyright (C) 2018 Peijun Ma
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path
from subprocess import DEVNULL, PIPE, run


def format_time(seconds):
    """
    Format time in secods into human readable string
    Args:
        seconds: time in seconds
    Returns:
        The time formatted
    """
    mins, secs = divmod(seconds, 60)
    return '{}:{:02d}'.format(int(mins), round(secs))


def get_supported_formats(ffplay):
    """
    Get formats supported by `ffplay`
    Args:
        ffplay: Path to ffplay binary

    Returns:
        Generator of supported formats
    """
    proc = run([ffplay, '-formats'], stderr=DEVNULL, stdout=PIPE)
    for line in proc.stdout.decode().splitlines():
        line = line.strip().split()
        if len(line) >= 3 and line[0].startswith('D'):
            yield from line[1].split(',')


def get_songs(path: Path, formats):
    """
    Recursively get songs under path
    Args:
        path: path to the music directory
        formats: set of supported audio formats
    Returns:
        Generator of all song paths under path
    """
    for sub in path.iterdir():
        if sub.is_file():
            if sub.suffix.lstrip('.') in formats:
                yield sub.absolute()
        elif sub.is_dir():
            yield from get_songs(sub, formats)



#  musicview, (re)discover your music library.
#  Copyright (C) 2018 Peijun Ma
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from itertools import chain
from pathlib import Path
from sqlite3 import Connection, connect
from typing import Tuple

import click
from click import progressbar

from .misc import get_songs, get_supported_formats
from .song import MetaData, song_metadata


def update_db(path: Path, conn: Connection, ffmpeg: str, ffplay: str):
    """
    Update the database
    Args:
        path: path to the music library
        conn: database connection
        ffmpeg: ffmpeg binary
        ffplay: ffplay binary
    """
    formats = set(get_supported_formats(ffplay))
    cur = conn.cursor()
    click.echo('Fetching songs...')
    songs = set(map(str, get_songs(path, formats)))
    cur.execute('CREATE TEMP TABLE tmp (path VARCHAR PRIMARY KEY);')
    cur.executemany('INSERT INTO tmp(path) VALUES (?);', ((s,) for s in songs))
    cur.execute('DELETE FROM library WHERE path NOT IN (SELECT path FROM tmp);')
    cur.execute('DROP TABLE tmp;')
    with progressbar(songs, label='Updating database...') as bar:
        for song in bar:
            metadata = song_metadata(ffmpeg, song)
            if not metadata:
                continue
            assert metadata.length
            cur.execute('SELECT * FROM library WHERE path=?', (metadata.path,))
            if not cur.fetchone():
                cur.execute(
                    """
                    INSERT INTO library (path, title, genre, artist, album, length)
                    VALUES (?, ?, ?, ?, ?, ?);
                    """, metadata
                )
            else:
                cur.execute(
                    """
                    UPDATE library SET
                    title = ?,
                    genre = ?,
                    artist = ?,
                    album = ?,
                    length = ?
                    WHERE path=?;
                    """, tuple(chain(metadata[1:], metadata[:1]))
                )
        conn.commit()
        cur.close()


def init_db(path: Path, db_path: Path, ffmpeg: str, ffplay: str):
    """
    Initialize the database
    Args:
        path: path to the music directory
        db_path: path to the database
        ffmpeg: ffmpeg binary
        ffplay: ffplay binary
    """
    db_path.touch()

    with connect(str(db_path)) as conn:
        conn.execute(
            """
            CREATE TABLE IF NOT EXISTS library (
            path VARCHAR PRIMARY KEY,
            title VARCHAR,
            genre VARCHAR,
            artist VARCHAR,
            album VARCHAR,
            length REAL,
            favourite BOOLEAN DEFAULT 0 NOT NULL,
            listen_count INT DEFAULT 0
            );
            """
        )
        conn.commit()
        update_db(path, conn, ffmpeg, ffplay)


def next_song(conn) -> Tuple[MetaData, tuple]:
    """
    Select the next song to play based on play counts
    Args:
        conn: the database connection
    Returns:
        metadata of the next song to play
    """
    cur = conn.cursor()
    cur.execute(
        """
        SELECT * FROM library
        WHERE listen_count=(SELECT min(listen_count) FROM library)
        ORDER BY RANDOM() LIMIT 1
        """
    )
    next = cur.fetchone()
    cur.execute(
        """
        UPDATE library SET 
        listen_count = listen_count + 1
        WHERE path=?
        """, (next[0],)
    )
    conn.commit()
    cur.close()
    return MetaData(*next[:-2]), next[-2:]
